/* $USAGI: udpsend.c,v 1.32 2005/09/12 12:55:55 yoshfuji Exp $ */

/*
 * Copyright (C) 2001,2002,2003,2004 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>

#if defined(HAVE_CONFIG_H)
# include "../config.h"
#endif

#define	PACKET_STRING	"TestPacket"

struct ip6optbuf {
	struct msghdr msg;
	struct cmsghdr *cmsg;
};

#if defined(IPV6_HOPOPTS)
# if !defined(IPV6_RECVHOPOPTS)
#  define OPT_COMPAT_H	"-c "
# elif defined(IPV6_2292HOPOPTS)
#  define OPT_COMPAT_H	"[-c] "
# endif
#else
# define OPT_COMPAT_H
#endif

#if defined(IPV6_DSTOPTS)
# if !defined(IPV6_RECVDSTOPTS)
#  define OPT_COMPAT_D	"-c "
# elif defined(IPV6_2292DSTOPTS)
#  define OPT_COMPAT_D	"[-c] "
# endif
#else
# define OPT_COMPAT_D
#endif

#if defined(IPV6_RTHDR)
# if !defined(IPV6_RECVRTHDR)
#  define OPT_COMPAT_R	"-c "
# elif defined(IPV6_2292RTHDR)
#  define OPT_COMPAT_R	"[-c] "
# endif
#else
# define OPT_COMPAT_R
#endif

#ifdef IPV6_PKTOPTIONS
# define OPT_PKTOPT	"[-k] "
#else
# define OPT_PKTOPT
#endif

int sock_create(sa_family_t family, int socktype, int proto);

int setup_rthdr(struct cmsghdr *cmsg, int type, const char *addrs)
{
	unsigned char *rth;
	int optlen = 8, segleft = 0;
	struct in6_addr *rtaddrs;
	char *p, *p0, *buf;

	buf = strdup(addrs);
	if (!buf) {
		perror("strdup");
		exit(1);
	}

	cmsg->cmsg_level = IPPROTO_IPV6;
	cmsg->cmsg_type = type;

	rth = CMSG_DATA(cmsg);
	memset(rth, 0, 8);

	rtaddrs = (struct in6_addr *)(rth + 8);

	p = p0 = buf;
	for (;;) {
		if (*p == ',' || *p == '\0') {
			char c = *p;
			*p = '\0';
			if (inet_pton(AF_INET6, p0, &rtaddrs[segleft]) <= 0) {
				perror("inet_pton");
				exit(1);
			}
			segleft++;
			if (!c)
				break;
			p0 = p + 1;
		}
		p++;
	}

	optlen += sizeof(struct in6_addr) * segleft;

	rth[1] = (optlen - 1) << 3;
	rth[3] = segleft;

	cmsg->cmsg_len = CMSG_LEN(optlen);
}

int setup_ip6opt(struct cmsghdr *cmsg, int type, int optlen)
{
	unsigned char *buf, *p;
	if (optlen < 8 || optlen > 255 * 8 || optlen % 8) {
		fprintf(stderr, "invalid optlen: %d\n", optlen);
		return -1;
	}

	cmsg->cmsg_level = IPPROTO_IPV6;
	cmsg->cmsg_type = type;
	cmsg->cmsg_len = CMSG_LEN(optlen);

	p = CMSG_DATA(cmsg);

	memset(p, 0, optlen);

	p[1] = (optlen >> 3) - 1;

	p += 2;
	optlen -= 2;

	while (optlen > 0) {
		int len = optlen;

		if (len == 1) {
			/* Pad1 */
			p[0] = 0;
		} else {
			/* PadN */
			if (len > 257)
				len = 257;
			p[0] = 1;
			p[1] = len - 2;
		}

		p += len;
		optlen -= len;
	}

	return 0;
}

#ifdef IPV6_TCLASS
int setup_tclass(struct cmsghdr *cmsg, int tclass)
{
	int *p;

	if (tclass < 0 || tclass > 255) {
		fprintf(stderr, "invalid tclass: %d\n", tclass);
		return -1;
	}

	cmsg->cmsg_level = IPPROTO_IPV6;
	cmsg->cmsg_type = IPV6_TCLASS;
	cmsg->cmsg_len = CMSG_LEN(sizeof(tclass));

	p = (int *)CMSG_DATA(cmsg);
	memcpy(p, &tclass, sizeof(*p));

	return 0;
}
#endif

int flush_pktopt(int s, struct ip6optbuf *optbuf)
{
#if 0
	printf("flush_pktopt()\n");
#endif

	if (!s)
		return 0;

#ifdef IPV6_PKTOPTIONS
	if ((optbuf)->msg.msg_control &&
	    setsockopt((s), IPPROTO_IPV6, IPV6_PKTOPTIONS,
		       (optbuf)->msg.msg_control,
		       (optbuf)->msg.msg_controllen) < 0) {
		perror("setsockopt(IPV6_PKTOPTIONS)");
		return -1;
	}
#endif
	(optbuf)->cmsg = NULL;
	free((optbuf)->msg.msg_control);
	memset(&(optbuf)->msg, 0, sizeof((optbuf)->msg));

	return 0;
}

int main(int argc, char **argv) {
	const char *argv0 = argv[0];
	char *host, *serv;
	struct addrinfo hints, *res;
	int gai;
	int s = -1;
	int len;
	char *buf;
	int chunk_size = 100;
	int send_flags = 0;
	int pref_privacy = 0;
	int packet_size = 100;
	int c, usage = 0;
	int use_connect = 0;
	int send_dst = 0, send_hbh = 0;
	char *send_rthdr = NULL;
	sa_family_t hints_family = AF_UNSPEC;
	struct ip6optbuf optbuf[2];
	int in_pktopt = 0, opt_compat = 0;
	int tclass = 0;

	memset(&optbuf, 0, sizeof(optbuf));

	printf("Start parsing arguments\n");
	while((c = getopt(argc, argv, 
				"c"
				"C"
#ifdef IPV6_RTHDRDSTOPTS
				"D:"
#endif
#ifdef IPV6_DSTOPTS
				"d:"
#endif
#ifdef IPV6_HOPOPTS
				"h:"
#endif
				"k"
#if defined(MSG_MORE)
				"m:"
#endif
				"p:"
#ifdef IPV6_DSTOPTS
				"r:"
#endif
				"s:"
#ifdef IPV6_TCLASS
				"t:"
#endif
			  )) != EOF){
		switch (c) {
		case 'c':
			flush_pktopt(s, &optbuf[1]);
			opt_compat = 1;
			break;
		case 'C':
			flush_pktopt(s, &optbuf[1]);
			opt_compat = 0;
			in_pktopt = 0;
			use_connect = 1;
			break;
		case 'k':
			in_pktopt = 1;
			break;
#if defined(IPV6_RTHDRDSTOPTS) || \
    defined(IPV6_DSTOPTS) || \
    defined(IPV6_HOPOPTS) || \
    defined(IPV6_RTHDR) || \
    defined(IPV6_TCLASS)
#ifdef IPV6_RTHDRDSTOPTS
		case 'D':
#endif
#ifdef IPV6_DSTOPTS
		case 'd':
#endif
#ifdef IPV6_HOPOPTS
		case 'h':
#endif
#ifdef IPV6_RTHDR
		case 'r':
#endif
#ifdef IPV6_TCLASS
		case 't':
#endif
		    {
			int optname = -1;
			int len;
			unsigned char *p;

#if 0
			printf("parsing -%c, %d\n", c, in_pktopt);
#endif
			if (s < 0) {
				s = sock_create(AF_INET6, SOCK_DGRAM, IPPROTO_UDP);
				hints_family = AF_INET6;
			}

			if ((!in_pktopt || !opt_compat) && optbuf[1].cmsg)
				flush_pktopt(s, &optbuf[1]);

			switch(c) {
#ifdef IPV6_RTHDRDSTOPTS
			case 'D':
				if (opt_compat) {
					usage = 1;
					break;
				}
				optname = IPV6_RTHDRDSTOPTS;
				len = atoi(optarg);
				break;
#endif
#ifdef IPV6_TCLASS
			case 't':
				if (opt_compat) {
					usage = 1;
					break;
				}
				optname = IPV6_TCLASS;
				tclass = atoi(optarg);
				len = sizeof(int);
				break;
#endif
#ifdef IPV6_DSTOPTS
			case 'd':
#ifdef IPV6_2292DSTOPTS
				optname = opt_compat ? IPV6_2292DSTOPTS : IPV6_DSTOPTS;
#else
				optname = IPV6_DSTOPTS;
#endif
				len = atoi(optarg);
				break;
#endif
#ifdef IPV6_HOPOPTS
			case 'h':
#ifdef IPV6_2292HOPOPTS
				optname = opt_compat ? IPV6_2292HOPOPTS : IPV6_HOPOPTS;
#else
				optname = IPV6_HOPOPTS;
#endif
				len = atoi(optarg);
				break;
#endif
#ifdef IPV6_RTHDR
			case 'r':
			    {
				int count = 1;
				for (p = optarg; *p; p++) {
					if (*p == ',')
						count++;
				}
				if (count > 127 || sizeof(struct in6_addr) * count > 255 * 8) {
					/* Too many hops */
					exit(1);
				}
				len = 8 + sizeof(struct in6_addr) * count;
#ifdef IPV6_2292RTHDR
				optname = opt_compat ? IPV6_2292RTHDR : IPV6_RTHDR;
#else
				optname = IPV6_RTHDR;
#endif
				break;
			    }
#endif
			}

			if (optname < 0)
				break;

			p = realloc(optbuf[in_pktopt].msg.msg_control,
				    optbuf[in_pktopt].msg.msg_controllen + CMSG_SPACE(len));
			if (!p)
				perror("realloc");

			/*
			 * NOTE: Because CMSG_NXTHDR() implementation of glibc
			 * has extra check about cmsg_len in next cmsg,
			 * we need to clear that field first.
			 */
			memset(p + optbuf[in_pktopt].msg.msg_controllen, 0, CMSG_SPACE(len));

			optbuf[in_pktopt].msg.msg_controllen += CMSG_SPACE(len);

			if (optbuf[in_pktopt].cmsg) {
				optbuf[in_pktopt].cmsg = (struct cmsghdr *)(((char *)p) +
									    ((char *)optbuf[in_pktopt].cmsg -
									     (char *)optbuf[in_pktopt].msg.msg_control));
				optbuf[in_pktopt].msg.msg_control = p;
#if 0
				optbuf[in_pktopt].cmsg = (struct cmsghdr *) ((unsigned char *) optbuf[in_pktopt].cmsg +
									     CMSG_ALIGN (optbuf[in_pktopt].cmsg->cmsg_len));
#else
				optbuf[in_pktopt].cmsg = CMSG_NXTHDR(&optbuf[in_pktopt].msg, optbuf[in_pktopt].cmsg);
#endif
			} else {
				optbuf[in_pktopt].msg.msg_control = p;
				optbuf[in_pktopt].cmsg = CMSG_FIRSTHDR(&optbuf[in_pktopt].msg);
			}

			switch(optname){
#ifdef IPV6_RTHDR
			case IPV6_RTHDR:
#ifdef IPV6_2292RTHDR
			case IPV6_2292RTHDR:
#endif
				if (setup_rthdr(optbuf[in_pktopt].cmsg, optname, optarg) < 0)
					exit(1);
				break;
#endif
#ifdef IPV6_TCLASS
			case IPV6_TCLASS:
				if (setup_tclass(optbuf[in_pktopt].cmsg, tclass) < 0)
					exit(1);
				break;
#endif
			default:
				if (setup_ip6opt(optbuf[in_pktopt].cmsg, optname, len) < 0)
					exit(1);
				break;
			}
			if (!opt_compat && in_pktopt) {
				assert(optbuf[in_pktopt].cmsg != NULL);
				printf("setsockopt(s=%d,type=%d,optname=%d,len=%d)\n", s, IPPROTO_IPV6, optname,len);
				if (setsockopt(s, IPPROTO_IPV6, optname,
				       CMSG_DATA(optbuf[in_pktopt].cmsg),
				       len) < 0) {
					perror("setsockopt()");
				}
				optbuf[in_pktopt].cmsg = NULL;
				free(optbuf[in_pktopt].msg.msg_control);
				memset(&optbuf[in_pktopt].msg, 0, sizeof(optbuf[in_pktopt].msg));
			}
			in_pktopt = 0;
			opt_compat = 0;
			break;
		    }
#endif
#if defined(MSG_MORE)
		case 'm':
			flush_pktopt(s, &optbuf[1]);
			opt_compat = 0;
			in_pktopt = 0;
			send_flags |= MSG_MORE;
			chunk_size = atoi(optarg);
			break;
#endif
		case 'p':
			flush_pktopt(s, &optbuf[1]);
			opt_compat = 0;
			in_pktopt = 0;
			if (s < 0) {
				s = sock_create(AF_INET6, SOCK_DGRAM, IPPROTO_UDP);
				hints_family = AF_INET6;
			}
			pref_privacy = atoi(optarg);
			break;
		case 's':
			flush_pktopt(s, &optbuf[1]);
			opt_compat = 0;
			in_pktopt = 0;
			packet_size = atoi(optarg);
			break;
		default:
			flush_pktopt(s, &optbuf[1]);
			opt_compat = 0;
			in_pktopt = 0;
			usage = 1;
			break;
		}
	}

	printf("Finished parsing arguments\n");

	if (optbuf[1].cmsg)
		flush_pktopt(s, &optbuf[1]);

	argc -= optind;
	argv += optind;

	if (usage || argc < 2 || chunk_size <= 0 || packet_size <= 0) {
		printf("%s "
			"[-C] "
#ifdef IPV6_RTHDRDSTOPTS
			"[" OPT_PKTOPT "-D dstopt_size] "
#endif
#ifdef IPV6_DSTOPTS
			"[" OPT_COMPAT_D OPT_PKTOPT "-d dstopt_size] "
#endif
#ifdef IPV6_HOPOPTS
			"[" OPT_COMPAT_H OPT_PKTOPT "-h hbhopt_size] "
#endif
#if defined(MSG_MORE)
			"[-m chunk_size] "
#endif
			"[-p pref_privacy] "
#ifdef IPV6_RTHDR
			"[" OPT_COMPAT_R OPT_PKTOPT "-r addr,addr,...] "
#endif
			"[-s packet_size] "
			"host port\n"
			"\n"
#if defined(IPV6_DSTOPTS)
# if defined(IPV6_RECVDSTOPTS)
			"\t      -d: send IPV6_DSTOPTS [3542] via ancillary data\n"
			"\t   -k -d: send IPV6_DSTOPTS [3542] socket option\n"
#  if defined(IPV6_2292DSTOPTS)
			"\t-c    -d: send IPV6_2292DSTOPTS (IPV6_DSTOPTS [2292]) via ancillary data\n"
#  endif
# else
			"\t-c    -d: send IPV6_DSTOPTS [2292] via ancillary data\n"
# endif
# if defined(IPV6_PKTOPTIONS)
			"\t-c -k -d: send IPV6_DSTOPTS [2292] via IPV6_PKTOPTIONS [2292] socket option\n"
# elif defined(IPV6_2292DSTOPTS) && defined(IPV6_2292PKTOPTIONS)
			"\t-c -k -d: send IPV6_2292DSTOPTS [2292] via IPV6_2292PKTOPTIONS [2292] socket option\n"
# endif
#endif

#if defined(IPV6_RTHDRDSTOPT)
# if defined(IPV6_RECVRTHDRDSTOPT)
			"\t-D      : send IPV6_RTHDRDSTOPTS [3542] via ancillary data\n"
			"\t   -k -D: send IPV6_RTHDRDSTOPTS [3542] socket option\n"
#  if defined(IPV6_2292RTHDRDSTOPT)
			"\t-c    -D: send IPV6_2292RTHDRDSTOPT (IPV6_RTHDRDSTOPT [2292]) via ancillary data\n"
#  endif
# else
			"\t-c    -D: send IPV6_RTHDRDSTOPT [2292] via ancillary data\n"
# endif
# if defined(IPV6_PKTOPTIONS)
			"\t-c -k -D: send IPV6_RTHDRDSTOPT [2292] via IPV6_PKTOPTIONS [2292] socket option\n"
# elif defined(IPV6_2292RTHDRDSTOPT) && defined(IPV6_2292PKTOPTIONS)
			"\t-c -k -D: send IPV6_2292RTHDRDSTOPT [2292] via IPV6_2292PKTOPTIONS [2292] socket option\n"
# endif
#endif

#if defined(IPV6_HOPOPTS)
# if defined(IPV6_RECVHOPOPTS)
			"\t      -h: send IPV6_HOPOPTS [3542] via ancillary data\n"
			"\t   -k -h: send IPV6_HOPOPTS [3542] socket option\n"
#  if defined(IPV6_2292HOPOPTS)
			"\t-c    -h: send IPV6_2292HOPOPTS (IPV6_HOPOPTS [2292]) via ancillary data\n"
#  endif
# else
			"\t-c    -h: send IPV6_HOPOPTS [2292] via ancillary data\n"
# endif
# if defined(IPV6_PKTOPTIONS)
			"\t-c -k -h: send IPV6_HOPOPTS [2292] via IPV6_PKTOPTIONS [2292] socket option\n"
# elif defined(IPV6_2292HOPOPTS) && defined(IPV6_2292PKTOPTIONS)
			"\t-c -k -h: send IPV6_2292HOPOPTS [2292] via IPV6_2292PKTOPTIONS [2292] socket option\n"
# endif
#endif

#if defined(IPV6_RTHDR)
# if defined(IPV6_RECVRTHDR)
			"\t      -r: send IPV6_RTHDR [3542] via ancillary data\n"
			"\t   -k -r: send IPV6_RTHDR [3542] socket option\n"
#  if defined(IPV6_2292RTHDR)
			"\t-c    -r: send IPV6_2292RTHDR (IPV6_RTHDR [2292]) via ancillary data\n"
#  endif
# else
			"\t-c    -r: send IPV6_RTHDR [2292] via ancillary data\n"
# endif
# if defined(IPV6_PKTOPTIONS)
			"\t-c -k -r: send IPV6_RTHDR [2292] via IPV6_PKTOPTIONS [2292] socket option\n"
# elif defined(IPV6_2292RTHDR) && defined(IPV6_2292PKTOPTIONS)
			"\t-c -k -r: send IPV6_2292RTHDR [2292] via IPV6_2292PKTOPTIONS [2292] socket option\n"
# endif
#endif

#if defined(IPV6_TCLASS)
			"\t      -t: send IPV6_TCLASS [3542] via ancillary data\n"
			"\t   -k -t: send IPV6_TCLASS [3542] socket option\n"
#endif
			"" , argv0);
		exit(1);
	}
	host = argv[0];
	serv = argv[1];

#if defined(MSG_MORE)
	if (!(send_flags & MSG_MORE))
		chunk_size = packet_size;
#else
	chunk_size = packet_size;
#endif
	buf = malloc(chunk_size);
	if (!buf) {
		perror("malloc");
		exit(1);
	}
	memset(buf, 0, chunk_size);
	snprintf(buf, chunk_size, "%s", PACKET_STRING);

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = hints_family;
	hints.ai_socktype = SOCK_DGRAM;
	gai = getaddrinfo(host, serv, &hints, &res);
	if (gai) {
		printf("getaddrinfo() for %s port %s: %s\n",
			host, serv, gai_strerror(gai));
		exit(1);
	}

	if (s < 0)
		s = sock_create(res->ai_family, res->ai_socktype, res->ai_protocol);
	else if (res->ai_family != hints_family) {
		errx(1, "getaddrinfo() returns inappropriate family %d (!= %d)",
		     (int)res->ai_family, (int)hints_family);
	}

	if (res->ai_family == PF_INET6 && pref_privacy) {
#ifdef IPV6_PRIVACY
		if (setsockopt(s, IPPROTO_IPV6, IPV6_PRIVACY, &pref_privacy, sizeof(pref_privacy)) < 0)
			perror("setsockopt(IPV6_PRIVACY)");
#else
		printf("IPV6_PRIVACY is not supported.\n");
#endif
	}	

	if (use_connect) {
		if (connect(s, res->ai_addr, res->ai_addrlen) < 0) {
			perror("connect");
			close(s);
			exit(1);
		}
	} else {
		optbuf[0].msg.msg_name = res->ai_addr;
		optbuf[0].msg.msg_namelen = res->ai_addrlen;
	}

	while (packet_size > 0) {
		struct iovec iov;

		if (packet_size <= chunk_size) {
			chunk_size = packet_size;
#if defined(MSG_MORE)
			send_flags &= ~MSG_MORE;
#endif
		}

		iov.iov_base = buf;
		iov.iov_len = chunk_size;

		optbuf[0].msg.msg_iov = &iov;
		optbuf[0].msg.msg_iovlen = 1;
#if 1
		len = sendmsg(s, &optbuf[0].msg, send_flags);
		if (len < 0) {
			perror("sendmsg");
			close(s);
			exit(1);
		}
#else
		if (use_connect) {
			if ((len = send(s, buf, chunk_size, send_flags)) < 0) {
				perror("send");
				close(s);
				exit(1);
			}
		} else {
			if ((len = sendto(s, buf, chunk_size, send_flags,
						res->ai_addr, res->ai_addrlen)) < 0) {
				perror("sendto");
				close(s);
				exit(1);
			}
		}
#endif
		printf ("len = %d, flags = %08x, sent = %d\n", chunk_size, send_flags, len);
		packet_size -= chunk_size;
	}

	close(s);
	exit(0);
}

int sock_create(sa_family_t family, int socktype, int proto)
{
	int s;

	printf("Creating socket (family = %d) ...", family);
	s = socket(family, socktype, proto);
	if (s < 0) {
		perror("socket");
		exit(2);
	}
	printf(" %d\n", s);
	return s;
}

